"""Declares :class:`PayloadCodec`."""
import typing
from typing import Any

from ckms.core import Keychain
from ckms.core.models import JSONWebSignature
from ckms.types import ClaimSet
from ckms.types import Data
from ckms.types import Encrypter
from ckms.types import IKeychain
from ckms.types import IKeySpecification
from ckms.types import JSONWebToken
from ckms.types import Signer
from .decoder import Decoder
from .encoder import Encoder
from .joseheaderset import JOSEHeaderSet
from .josepayload import JOSEPayload
from .validator import Validator


SUPPORTED_ALGORITHMS = [
    "RSA1_5",
    "RSA-OAEP",
    "RSA-OAEP-256",
    "RSA-OAEP-384",
    "RSA-OAEP-512",
    "A128KW",
    "A192KW",
    "A256KW",
    "A128GCMKW",
    "A192GCMKW",
    "A256GCMKW",
    "dir",
    "ECDH-ES",
    "ECDH-ES+A128KW",
    "ECDH-ES+A192KW",
    "ECDH-ES+A256KW",
    "PBES2-HS256+A128KW",
    "PBES2-HS256+A192KW",
    "PBES2-HS256+A256KW",
]


class PayloadCodec:
    """Exposes an interface to decode and encode a JSON Web Token (JWT),
    JSON Web Encryption (JWE) or JSON Web Signature (JWS).
    """
    __module__: str = 'ckms.jose'
    allow_compact: bool
    allow_flattened: bool
    decoder: Decoder
    encoder: Encoder
    encryption_keys: list[str]
    signer: IKeychain
    verifier: IKeychain
    decrypter: IKeychain
    encrypter: IKeychain
    allow_jwks_import: bool
    signing_keys: list [str | IKeySpecification]
    trusted_issuers: set[str]
    validator: Validator | None

    def __init__(
        self,
        signer: IKeychain | None = None,
        verifier: IKeychain | None = None,
        decrypter: IKeychain | None = None,
        encrypter: IKeychain | None = None,
        required_claims: typing.Optional[typing.Set[str]] = None,
        jwks_import: bool = False,
        audiences: typing.Optional[typing.Set[str]] = None,
        trusted_issuers: typing.Optional[typing.Set[str]] = None,
        content_encryption: str = "A256GCM",
        signing_keys: list[str | IKeySpecification] | None = None,
        encryption_keys: list[str | IKeySpecification] | None = None,
        allow_compact: bool = True,
        allow_flattened: bool = True
    ):
        """A :class:`PayloadCodec` instance governs the rules under which
        objects are deserialized, decoded and instantiated into the
        corresponding Python objects.

        The `verifier` and `decrypter` parameters are required and specify
        which :class:`~ckms.types.IKeychain` implementations are used to verify
        digital signatures and decrypt JWE payloads. A common configuration is
        to provide a :class:`~ckms.types.JSONWebKeySet` holding public keys
        exclusively as the `verifier` parameter, and a :class:`ckms.core.Keychain`
        or similar as the `decrypter` parameter. If any of these parameters
        is omitted, then the respective operations can not be performed by
        :class:`PayloadCodec` and will if the (public) keys can not be
        resolved otherwise (e.g. through the issuer JWKS URI for example).

        A provider of entropy may be specified using the `random` argument.

        The `required_claims` specifies the claims the must be present when
        the payload is a JWT. If any of these claims is missing, then
        validation of the JWT fails. Null values are not considered missing
        claims.

        The `jwks_import` parameter is a boolean indicating if a JSON
        Web Key Set (JWKS) or JSON Web Key (JWK) may be imported using either
        the ``iss`` (using OAuth 2.0 metadata discovery) or ``jku`` claims,
        respectively. The default is ``False``.

        With the `trusted_issuers` parameter the set of issuers that are
        trusted by the codec may be specified. Any token that has an ``iss``
        claim that is not in this set, is rejected. If `jwks_import``
        is ``True``, then a JWK(S) may be imported if the issuer is in
        this set.  If `trusted_issuers` is ``None`` or an empty set, then the
        `iss` claim is not considered during JWT validation.

        Similarly, the `audiences` parameter specifies the set of audiences
        with which the set of audiences in the `aud` claim of a JWT must
        intersect. If `audiences` is ``None`` or an empty set, then the
        `aud` claim is not considered during JWT validation.

        Default signers may be configured with the `signing_keys` parameter,
        which is a list of strings specifying the keys to sign a JWS with.
        The signatures are created in order. Likewise, the `encryption_keys`
        parameter is used to specify default encryption keys.
        """
        self.signer = signer if signer is not None else Keychain()
        self.verifier = verifier if verifier is not None else Keychain()
        self.decrypter = decrypter if decrypter is not None else Keychain()
        self.encrypter = encrypter if encrypter is not None else Keychain()
        self.decoder = Decoder(decrypter=self.decrypter, verifier=self.verifier)
        self.encoder = Encoder(
            allow_compact=allow_compact,
            allow_flattened=allow_flattened,
            encrypter=self.encrypter,
            encryption_keys=encryption_keys,
            signer=self.signer,
            signing_keys=signing_keys
        )
        self.jwks_import = jwks_import
        self.signing_keys = signing_keys or []
        self.trusted_issuers = set(trusted_issuers or [])
        self.issuers = {}
        self.validator = None
        self.content_encryption = content_encryption

    async def decode(
        self,
        token: bytes | str,
        verify: bool = True,
        accept: set[str] = {"at+jwt", "jwt", "secevent+jwt"},
        require_kid: bool = True,
        _type: type = JSONWebToken
    ) -> bytes | str | JSONWebToken:
        """Decode `token` and return the payload, which is either a
        :class:`~ckms.types.JSONWebToken` or byte-sequence. See also
        :meth:`~ckms.jose.Decoder.decode()`.
        """
        return await self.decoder.decode(
            token,
            verify=verify,
            accept=accept,
            require_kid=require_kid
        )

    async def jws(self, token: bytes | str) -> JSONWebSignature:
        """Like :meth:`decode()`, but returns the unverified and unparsed
        JSON Web Signature (JWS). If `token` is not a JWS, raise an
        exception.
        """
        return await self.decoder.jws(token)

    async def jwt(
        self,
        token: bytes | str,
        accept: set[str] | str | None = None
    ) -> tuple[JSONWebSignature, JSONWebToken]:
        """Like :meth:`decode()`, but returns a tuple containing the
        unverified JSON Web Signature (JWS) and its claims set. If
        `token` is not a JWS, or does not have a claims set, raise
        an exception.
        """
        return await self.decoder.jwt(token, accept=accept)

    def encode(
        self,
        payload: bytes | dict[str, Any] | str | ClaimSet | Data,
        sign: bool = True,
        encrypt: bool = True,
        content_type: str | None = None,
        claimset_class: type[ClaimSet] = ClaimSet,
        signing_keys: list[str] | None = None,
        signers: list[str] | list[Signer] | None = None,
        encrypters: list[str] | list[Encrypter] | None = None,
        allow_compact: bool = True,
        allow_flattened: bool = True
    ) -> JOSEPayload:
        return self.encoder.encode(
            content=payload,
            sign=sign,
            encrypt=encrypt,
            content_type=content_type,
            claimset_class=claimset_class,
            signing_keys=signing_keys,
            signers=signers,
            encrypters=encrypters,
            allow_compact=True,
            allow_flattened=True
        )

    def introspect(
        self,
        token: str | bytes,
        model: type[JSONWebToken] = JSONWebToken,
        accept: set[str] = {"at+jwt", "jwt"}
    ) -> tuple[JOSEHeaderSet, ClaimSet | None]:
        """Introspect the token. Return a :class:`ckms.jose.JOSEHeaderSet`
        instance, and a :class:`ckms.types.JSONWebToken` instance representing the
        claims if the token is a JSON Web Token (JWT), otherwise the second
        variable in the tuple is byte-sequence enclosed in the object.
        """
        return self.decoder.introspect(token, model=model)

    def get_key_algorithm(self, using: str) -> str | None:
        """Return the algorithm for the specified key `using`."""
        return self.signer.get(using).algorithm



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum


class KeyOperationType(str, enum.Enum):
    decrypt     = 'decrypt'
    deriveBits  = 'deriveBits'
    deriveKey   = 'deriveKey'
    encrypt     = 'encrypt'
    sign        = 'sign'
    verify      = 'verify' 
    wrapKey     = 'wrapKey'
    unwrapKey   = 'unwrapKey'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from pyasn1.type import univ


class ObjectIdentifier(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: str | univ.ObjectIdentifier | None
    ) -> Optional['ObjectIdentifier']:
        if isinstance(value, univ.ObjectIdentifier):
            value = str(value) if value.isValue else None
        return cls(value) if value is not None else None

    def asn1(self) -> univ.ObjectIdentifier:
        return univ.ObjectIdentifier(self)


class Integer(int):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: int | univ.Integer
    ) -> 'Integer':
        if isinstance(value, univ.Integer):
            value = int(value)
        return cls(value)

    def asn1(self) -> univ.Integer:
        return univ.Integer(self)


class BitString(bytes):
    null: bool

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: bytes | univ.BitString | None
    ) -> Optional['BitString']:
        if isinstance(value, univ.BitString):
            value = bytes(value.asOctets()) if value.isValue else None # type: ignore
        return cls(value, null=value is None)

    def __new__(cls, value: bytes | None, null: bool = False):
        self = super().__new__(cls, value or b'')
        self.null = null
        return self

    def asn1(self) -> univ.BitString:
        return univ.BitString.fromOctetString(self) # type: ignore

    def __repr__(self) -> str:
        return bytes.__repr__(self)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import Any

import pytest

from ckms.core import Keychain
from ckms.core import KeyInspector
from ckms.core.models import KeySpecification
from ckms.utils import b64encode
from ckms.jose.encoder import Encoder


class JOSEProducerTestCase:
    __module__: str = 'ckms.jose.tests'
    inspector: KeyInspector = KeyInspector()
    supports_compact: bool = True
    supports_flattened: bool = True
    supports_json: bool = True
    supports_multi: bool = True

    def get_data(self) -> Any:
        return "Hello world!"

    def jwk_from_spec(self, spec: KeySpecification) -> dict[str, Any]:
        k: dict[str, Any]
        if spec.is_symmetric():
            k = dict(
                kty='oct',
                k=b64encode(spec.get_key_material()).decode('ascii')
            )
        else:
            k = dict(**spec.as_jwk().dict(exclude_defaults=True))
        return k

    async def create_jws(
        self,
        encoder: Encoder,
        signers: list[tuple[KeySpecification, Any, Any]],
        payload: bytes | dict[str, Any],
        compact: bool = False,
    ) -> Any:
        obj = encoder.encode(content=payload)
        for signer, protected, header in signers:
            assert not protected
            obj.add_signature(
                signer=signer,
                protected=protected,
                header=header
            )
        return await obj.serialize(compact=compact)

    async def verify_jws(
        self,
        verifiers: list[KeySpecification],
        token: bytes,
        payload: str | None = None,
        expect_jwt: bool = False
    ) -> None:
        raise NotImplementedError

    @pytest.mark.parametrize("compact", [False, True])
    @pytest.mark.parametrize("data", [None, {'foo': 'bar'}])
    async def test_sign_single_jws(
        self,
        keychain: Keychain,
        name: str,
        encoder: Encoder,
        compact: bool,
        data: Any
    ):
        """Encode a JWS with a single signature."""
        if compact and not self.supports_compact:
            pytest.skip()
        if not compact and not (self.supports_flattened or self.supports_multi):
            pytest.skip()
        spec = keychain.get(name)
        assert await self.verify_jws(
            verifiers=[spec],
            token=await self.create_jws(
                encoder=encoder,
                payload=data or self.get_data(),
                signers=[(spec, None, None)],
                compact=compact
            ),
            payload=json.dumps(data) if data else self.get_data(),
            expect_jwt=isinstance(data, dict)
        )

    async def test_sign_single_jws_different_key_does_not_verify(
        self,
        keychain: Keychain,
        name: str,
        hs256: KeySpecification,
        encoder: Encoder
    ):
        """Encode a JWS with a single signature."""
        spec = keychain.get(name)
        assert not await self.verify_jws(
            verifiers=[hs256],
            token=await self.create_jws(
                encoder=encoder,
                payload=self.get_data(),
                signers=[(spec, None, None)],
                compact=True
            ),
            payload=self.get_data()
        )

    @pytest.mark.parametrize("data", [None, {'foo': 'bar'}])
    async def test_sign_multi_jws(
        self,
        keychain: Keychain,
        name: str,
        hs256: KeySpecification,
        encoder: Encoder,
        data: Any
    ):
        if not self.supports_multi:
            pytest.skip()
        spec = keychain.get(name)
        assert await self.verify_jws(
            verifiers=[spec, hs256],
            token=await self.create_jws(
                encoder=encoder,
                payload=data or self.get_data(),
                signers=[
                    (spec, None, None),
                    (hs256, None, None)
                ],
                compact=False
            ),
            payload=json.dumps(data) if data else self.get_data()
        )

    async def test_sign_multi_jws_different_key_does_not_verify(
        self,
        keychain: Keychain,
        name: str,
        hs256: KeySpecification,
        hs384: KeySpecification,
        encoder: Encoder
    ):
        if not self.supports_multi:
            pytest.skip()
        spec = keychain.get(name)
        assert not await self.verify_jws(
            verifiers=[hs384],
            token=await self.create_jws(
                encoder=encoder,
                payload=self.get_data(),
                signers=[
                    (spec, None, None),
                    (hs256, None, None)
                ],
                compact=False
            ),
            payload=self.get_data()
        )


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Literal

from ckms import types
from ckms.core import models
from ckms.core import Provider


class LocalKeySpecification(models.KeySpecification):
    kty: str | None = None # type: ignore
    provider: Literal['local'] | Provider = 'local' # type: ignore
    key: Any

    @classmethod
    def autodiscover(
        cls,
        provider: types.IProvider,
        inspector: types.IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        values.update({
            'use': inspector.get_algorithm_use(values['algorithm'])
        })

    async def load(self) -> models.KeySpecification:
        """Load or generate the key as specified by the parameters."""
        assert isinstance(self.provider, types.IProvider)
        if not self.loaded:
            self.key = await self.key.setup(self)
        self.kid = self.provider.get_key_identifier(self)
        self.loaded = True
        return self

    def get_key_material(self) -> bytes:
        assert isinstance(self.provider, types.IProvider)
        return self.provider.inspector.to_pem(self.key)

    def has_key_material(self) -> bool:
        return True

    def get_private_key(self) -> Any:
        return self.key


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyasn1.type import univ
from pyasn1.type.base import Asn1Type
from pyasn1_modules import rfc2985
from pyasn1_modules import rfc5280

from ckms.ext import asn1


OID_SCHEMA: dict[str, type[Asn1Type]] = {
    '1.2.840.113549.1.9.14': rfc2985.Extensions
}


class Attribute(asn1.ASN1Model):
    """An ``Attribute`` as defined in :rfc:`5280`."""
    __asn1spec__: type[rfc5280.Attribute] = rfc5280.Attribute
    type: asn1.fields.ObjectIdentifier
    values: univ.SetOf

    def decode(self, spec: Asn1Type | None = None) -> list[Asn1Type]:
        if spec is None and self.type in OID_SCHEMA:
            spec = OID_SCHEMA[self.type]()
        return [asn1.decode_der(x, spec) for x in self.values] # type: ignore


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from ckms.core.provider.tests import AsymmetricSigningTestCase
from ckms.core.provider.tests import EncryptionTestCase
from ckms.core.provider.tests import ProviderTestCase
from ckms.core.provider.tests import SymmetricSigningTestCase

from .conftest import ENCRYPTION_KEYS
from .conftest import SIGNING_KEYS_ASYMMETRIC
from .conftest import SIGNING_KEYS_SYMMETRIC


class TestAzureProviderFacilities(ProviderTestCase):
    name: str = 'google'


@pytest.mark.parametrize("spec", SIGNING_KEYS_ASYMMETRIC)
class TestAzureAsymmetricSign(AsymmetricSigningTestCase):
    name: str = 'google'


@pytest.mark.parametrize("spec", SIGNING_KEYS_SYMMETRIC)
class TestAzureSymmetricSign(SymmetricSigningTestCase):
    name: str = 'google'


@pytest.mark.parametrize("spec", ENCRYPTION_KEYS)
class TestAzureEncryption(EncryptionTestCase):
    name: str = 'google'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pytest

from ckms.jose import PayloadCodec
from ckms.core import Keychain
from ckms.core.models import KeySpecification
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload
from .conftest import ENCRYPTION_KEYS



@pytest.mark.parametrize("spec", ENCRYPTION_KEYS)
class TestProviderJWE:

    @pytest.mark.parametrize("pt", [
        b"Hello world!",
        {'foo': 1}
    ])
    @pytest.mark.asyncio
    async def test_encrypt_and_decrypt(
        self,
        encrypter: Keychain,
        decrypter: Keychain,
        spec: KeySpecification,
        pt: str | dict[str, Any]
    ):
        c1 = PayloadCodec(decrypter=decrypter)
        c2 = PayloadCodec()
        jwe = await c1.encode(pt, encrypters=[spec])
        payload = await c1.decode(jwe)
        if isinstance(pt, bytes):
            assert pt == payload
        elif isinstance(pt, dict):
            assert isinstance(payload, JSONWebToken), jwe
        with pytest.raises(MalformedPayload):
            await c2.decode(jwe)


# pylint: skip-file
# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
from typing import cast
from typing import AsyncIterator

from azure.identity.aio import DefaultAzureCredential
from azure.keyvault.keys.aio import KeyClient
from azure.keyvault.keys.crypto.aio import CryptographyClient


class AzureConsumer:
    __module__: str = 'ckms.ext.azure.types'

    def credential(self) -> DefaultAzureCredential:
        return DefaultAzureCredential()

    @contextlib.asynccontextmanager
    async def crypto_client(self, key_id: str) -> AsyncIterator[CryptographyClient]:
        async with self.credential() as credential:
            async with CryptographyClient(key_id, credential=credential) as client: # type: ignore
                yield cast(CryptographyClient, client)

    @contextlib.asynccontextmanager
    async def key_client(self, vault_url: str) -> AsyncIterator[KeyClient]:
        async with self.credential() as credential:
            async with KeyClient(vault_url=vault_url, credential=credential) as client: # type: ignore
                yield cast(KeyClient, client)


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`JSONWebSignature`."""
import asyncio
import hashlib
from typing import Any
from typing import Literal

import pydantic

from ckms.types import IKeychain
from ckms.types import JSONWebToken
from ckms.types import MalformedHeader
from ckms.types import MalformedPayload
from ckms.utils import b64decode_json
from .octetb64 import OctetB64
from .joseheader_ import JOSEHeader
from .signature import Signature


class JSONWebSignature(pydantic.BaseModel):
    kind: Literal['JWS'] = pydantic.Field(..., exclude=True)
    payload: str
    signatures: list[Signature]

    @property
    def cty(self) -> str | None:
        """The type of JOSE object as defined by the ``cty`` claim."""
        return str.lower(self.header.cty) if self.header.cty else None

    @property
    def header(self) -> JOSEHeader:
        """The header of the first signature. The claims presented in this
        header are considered authoritative when determining the type,
        content type.
        """
        return self.signatures[0].header

    @property
    def typ(self) -> str | None:
        """The type of JOSE object as defined by the ``typ`` claim."""
        return str.lower(self.header.typ) if self.header.typ else None

    @pydantic.root_validator(pre=True, allow_reuse=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        for signature in (values.get('signatures') or []):
            protected = b64decode_json(
                signature.get('protected') or b'{}',
                on_failure=MalformedHeader(
                    detail=(
                        "The JWE Protected Header could not be interpreted as "
                        "a Base64-encoded JSON object."
                    )
                ),
                require=dict
            )

            assert isinstance(protected, dict)
            header = signature.get('header') or {}

            # The Header Parameter values used when creating or validating
            # individual signature or MAC values are the union of the two
            # sets of Header Parameter values that may be present: (1) the
            # JWS Protected Header represented in the "protected" member of
            # the signature/MAC's array element, and (2) the JWS Unprotected
            # Header in the "header" member of the signature/MAC's array element.
            # The union of these sets of Header Parameters comprises the JOSE
            # Header.  The Header Parameter names in the two locations MUST
            # be disjoint (RFC7515, 7.2.1).
            duplicates = set(protected.keys()) & set(header.keys())
            if duplicates:
                raise MalformedHeader(
                    detail=(
                        "The claims in a per-recipient header must be "
                        "disjoint from the protected and unprotected "
                        "headers."
                    ),
                    hint=(
                        "The following claims were duplicate: "
                        f"{', '.join(sorted(duplicates))}."
                    )
                )

            signature['header'] = {**header, **protected}

        return values

    def claims(self, accept: set[str]) -> JSONWebToken:
        """Return the claims set included in the JWS. If the JWS is not a JWT,
        raise an exception.
        """
        if self.typ not in accept:
            raise MalformedPayload(
                detail=(
                    "The JSON Web Signature (JWS) does not include a "
                    "claims set of an accepted type."
                )
            )
        return JSONWebToken.frompayload(self.payload)

    def digest(self, name: str = 'sha3_256') -> bytes:
        """Return a hash of the payload."""
        return hashlib.new(name, self.payload.encode('ascii')).digest()

    def hexdigest(self, *args: Any, **kwargs: Any) -> str:
        """Return a hex-encoded hash of the payload."""
        return self.digest().hex()

    def get_claims(self, model: type[JSONWebToken] = JSONWebToken):
        """Return the claims signed by the JWS."""
        return model.frompayload(self.payload)

    def get_headers(self) -> list[JOSEHeader]:
        """Return the list of :class:`JOSEHeader` containing the aggregated
        header claims for all signatures.
        """
        return [x.header for x in self.signatures]

    def get_payload(self) -> bytes:
        """Return the payload of te JWS."""
        return self.payload.encode('ascii')

    def get_type(self) -> str | None:
        """Return the normalized type of the JWS."""
        header = self.signatures[0].header
        return str.lower(header.typ or '') if header.typ else None

    def is_jwt(self, accept: set[str] = {"at+jwt", "jwt", "secevent+jwt"}) -> bool:
        """Return a boolean indicating if the :class:`JSONWebSignature` is
        a JSON Web Token (JWT).
        """
        return str.lower(self.signatures[0].header.typ or '') in accept

    async def verify(
        self,
        keychain: IKeychain,
        require_kid: bool = True
    ) -> bool:
        """Verify the JSON Web Signature (JWS) using the given
        :class:`~ckms.core.types.IKeychain` implementation
        `keychain`.
        """
        results = await asyncio.gather(*[
            x.verify(keychain, self.payload, require_kid=require_kid)
            for x in self.signatures
        ])
        return all(results)

    class Config:
        json_encoders: dict[Any, Any] = {
            bytes: lambda value: value.decode('ascii'), # type: ignore
            OctetB64: OctetB64.serialize
        }



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from typing import cast
from typing import Any
from typing import Awaitable
from typing import Callable

import crcmod
import google.auth
from google.api_core.exceptions import InvalidArgument
from google.auth.credentials import Credentials
from google.cloud.kms import AsymmetricDecryptResponse
from google.cloud.kms import DecryptResponse
from google.cloud.kms import KeyManagementServiceAsyncClient
from google.cloud.secretmanager import SecretManagerServiceAsyncClient

from ckms import core
from ckms.utils import normalize_ec_signature
from ckms.types import AESOperation
from ckms.types import CipherText
from ckms.types import EllipticCurveOperation
from ckms.types import HMACOperation
from ckms.types import RSAOperation
from ckms.types import SigningOperation
from .types import GoogleKeySpecification
from .types import IGoogleProvider
from .types import SecretManagerBlob


class GoogleProvider(IGoogleProvider):
    __module__: str = 'ckms.ext.google'
    credentials: Credentials
    project: str
    spec: type[GoogleKeySpecification] = GoogleKeySpecification
    timeout: int = 10.0

    @staticmethod
    def crc32c(data: bytes) -> int:
        """Calculates the CRC32C checksum of the provided data."""
        func = crcmod.predefined.mkPredefinedCrcFun('crc-32c')
        return func(data)

    def client(self) -> KeyManagementServiceAsyncClient:
        return KeyManagementServiceAsyncClient(credentials=self.credentials)

    def get_key_identifier(self, key: Any) -> str:
        return self.inspector.calculate_kid(key)

    def on_init(self):
        self.credentials, self.project = cast(Credentials, google.auth.default())

    async def decrypt_aes(self, op: AESOperation) -> bytes:
        spec = cast(GoogleKeySpecification, op.get_keyspec())
        async with self.client() as client:
            return await self.decrypt_remote(op, client.decrypt, spec.crypto_key)

    async def decrypt_remote(
        self,
        op: AESOperation | RSAOperation,
        func: Callable[..., Awaitable[AsymmetricDecryptResponse | DecryptResponse]],
        using: str
    ) -> bytes:
        spec = cast(GoogleKeySpecification, op.get_keyspec())
        ct = await op.get_message()
        request = {
            'name': using,
            'ciphertext': ct,
            'ciphertext_crc32c': self.crc32c(ct)
        }
        if op.aad is not None:
            assert spec.is_aead()
            request.update({
                'additional_authenticated_data': op.aad,
                'additional_authenticated_data_crc32c': self.crc32c(op.aad)
            })
        async with self.client() as client:
            try:
                response = await func(
                    request=request,
                    timeout=self.timeout
                )
            except InvalidArgument as exception:
                raise self.Undecryptable
            if not self.crc32c(response.plaintext) == response.plaintext_crc32c: # pragma: no cover
                raise ValueError("Corrupted plain text in transit.")

        return response.plaintext

    async def decrypt_rsa(self, op: RSAOperation) -> bytes:
        spec = cast(GoogleKeySpecification, op.get_keyspec())
        async with self.client() as client:
            return await self.decrypt_remote(op, client.asymmetric_decrypt, spec.resource_name)

    async def encrypt_aes(self, op: AESOperation) -> bytes:
        spec = cast(GoogleKeySpecification, op.get_keyspec())
        pt = await op.get_message()
        request = {
            'name': spec.resource_name,
            'plaintext': pt,
            'plaintext_crc32c': self.crc32c(pt)
        }
        assert spec.is_aead()
        if op.aad is not None:
            request.update({
                'additional_authenticated_data': op.aad,
                'additional_authenticated_data_crc32c': self.crc32c(op.aad)
            })
        async with self.client() as client:
            response = await client.encrypt(
                request=request,
                timeout=self.timeout
            )
            if not response.verified_plaintext_crc32c: # pragma: no cover
                raise ValueError("Plaintext checksum not considered or verified by cloud service.")
            if not response.verified_additional_authenticated_data_crc32c\
            and op.aad is not None: # pragma: no cover
                raise ValueError("AAD checksum not considered or verified by cloud service.")
            if not self.crc32c(response.ciphertext) == response.ciphertext_crc32c: # pragma: no cover
                raise ValueError("Corrupted cipher text in transit.")

        # Assume here that the Google Cloud KMS encryption key is never
        # used outside the HSM (i.e. if it was imported). Since tag/iv
        # values are expected, we fill them with zeroes here (TODO).
        return CipherText(
            aad=op.aad,
            buf=response.ciphertext,
            iv=b'0' * 12,
            tag=b'0' * 16
        )

    async def encrypt_rsa(self, op: RSAOperation) -> CipherText:
        p = self.get('local')
        return await p.encrypt_rsa(op)

    async def fetch(self, params: core.RemoteBlob) -> bytes: # pragma: no cover
        """Fetches the content of the remote blob using the parameters
        specified. Return a byte-string holding the content.
        """
        blob = SecretManagerBlob.parse_obj(params.dict())
        client = SecretManagerServiceAsyncClient(credentials=self.credentials)
        secret = client.secret_path(blob.project, blob.secret)
        async with client:
            response = await client.access_secret_version(
                request={
                    'name': f'{secret}/versions/{blob.version}'
                }
            )
        return response.payload.data

    async def sign_ec(self, op: EllipticCurveOperation) -> bytes:
        return await self.sign_asymmetric(op)

    async def sign_hmac(self, op: HMACOperation) -> bytes:
        return await self.sign_symmetric(op)

    async def sign_rsa(self, op: RSAOperation) -> bytes:
        return await self.sign_asymmetric(op)

    async def sign_symmetric(self, op: SigningOperation) -> bytes:
        spec = cast(GoogleKeySpecification, op.get_keyspec())
        data = await op.get_message()
        checksum = self.crc32c(data)
        request = {
            'name': spec.resource_name,
            'data': data,
            'data_crc32c': checksum
        }
        async with self.client() as client:
            response = await client.mac_sign(
                request=request,
                timeout=self.timeout
            )
            if not response.verified_data_crc32c: # pragma: no cover
                raise ValueError("Data checksum not considered or verified by cloud service.")
            if self.crc32c(response.mac) != response.mac_crc32c: # pragma: no cover
                raise ValueError("Corrupted MAC in transit.")
        return response.mac

    async def sign_asymmetric(self, op: SigningOperation) -> bytes:
        spec = cast(GoogleKeySpecification, op.get_keyspec())
        public = spec.get_public_key()
        data = await op.get_digest()
        checksum = self.crc32c(data)
        digest = {op.digest: data}
        request = {
            'name': spec.resource_name,
            'digest': digest,
            'digest_crc32c': checksum
        }
        async with self.client() as client:
            response = await client.asymmetric_sign(
                request=request,
                timeout=self.timeout
            )
            crc = response.signature_crc32c
            if not response.verified_digest_crc32c: # pragma: no cover
                raise ValueError("Digest checksum not considered or verified by cloud service.")
            if crc != self.crc32c(response.signature): # pragma: no cover
                raise ValueError("Corrupted signature in transit.")

        sig = response.signature
        if spec.kty == "EC":
            sig = normalize_ec_signature(
                l=(public.curve.key_size + 7) // 8,
                sig=sig
            )
        return sig

    async def verify_hmac(self, op: HMACOperation) -> bool:
        if op.signature is None: # pragma: no cover
            raise ValueError("Signature can not be None.")
        if len(op.signature) != 32:
            return False
        spec = cast(GoogleKeySpecification, op.get_keyspec())
        data = await op.get_message()
        checksum = self.crc32c(data)
        request = {
            'name': spec.resource_name,
            'data': data,
            'data_crc32c': checksum,
            'mac': op.signature,
            'mac_crc32c': self.crc32c(op.signature)
        }
        async with self.client() as client:
            response = await client.mac_verify(
                request=request,
                timeout=self.timeout
            )
            if not response.verified_mac_crc32c: # pragma: no cover
                raise ValueError("HMAC checksum not considered or verified by cloud service.")
            if not response.verified_data_crc32c: # pragma: no cover
                raise ValueError("Data checksum not considered or verified by cloud service.")
        return response.success


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`IHTTPClient`."""
from typing import Any
from typing import Awaitable

from .ihttpresponse import IHTTPResponse


class IHTTPClient:
    __module__: str = 'ckms.types'
    status_code: int

    def get(
        self,
        url: str,
        *args: Any,
        **kwargs: Any
    ) -> IHTTPResponse | Awaitable[IHTTPResponse]:
        """Perform a HTTP GET request."""
        raise NotImplementedError

    def post(
        self,
        url: str,
        *args: Any,
        **kwargs: Any
    ) -> IHTTPResponse | Awaitable[IHTTPResponse]:
        """Perform a HTTP POST request."""
        raise NotImplementedError


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Sized

import pytest

from ckms.core import Keychain
from ckms.core import const
from ckms.types import JSONWebKeySet


@pytest.fixture
def tagged_keys() -> dict[str, Any]:
    return {
        'HS256': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS256',
            'key': {'length': 32},
            'tags': {'foo'}
        },
        'HS384': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS384',
            'key': {'length': 32},
            'tags': {'bar'}
        },
        'HS512': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS512',
            'key': {'length': 32},
            'tags': {'foo', 'bar'}
        },
    }


@pytest.mark.asyncio
async def test_fromdict(keys: dict[str, Any]):
    kc = Keychain.fromdict(keys)
    await kc
    assert len(kc) == len(keys) * 2


def test_keychain_length(
    keychain: Keychain,
    keys: Sized
):
    assert len(keychain) == len(keys) * 2


@pytest.mark.parametrize("algorithm", const.ALGORITHMS)
def test_keychain_has(algorithm: str, keychain: Keychain):
    assert keychain.has(algorithm), algorithm


@pytest.mark.parametrize("algorithm", const.ALGORITHMS)
def test_add_duplicate_raises_valueerror(algorithm: str, keychain: Keychain):
    with pytest.raises(ValueError):
        keychain.add(algorithm, {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS256',
            'use': 'sig',
            'key': {'length': 32}
        })



@pytest.mark.parametrize("params", [
    {'kid': 'abc'},
    {'algorithm': 'foo'},
    {'use': 'bar'},
])
def test_filter_with_unloaded_raises_runtimeerror(params: dict[str, Any]):
    keychain = Keychain()
    with pytest.raises(RuntimeError):
        keychain.add('abc', {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS256',
            'use': 'sig',
            'key': {'length': 32}
        })
        keychain.filter(**params)


def test_filter_with_non_existing_kid_returns_empty(keychain: Keychain):
    result = keychain.filter(kid='abc')
    assert len(result) == 0


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS)
@pytest.mark.asyncio
async def test_asymmetric_key_is_not_aliased(
    algorithm: str,
    keychain: Keychain
):
    spec = keychain.get(algorithm)
    public = keychain.get(spec.kid) # type: ignore
    assert spec != public
    assert spec.kid == public.kid


@pytest.mark.parametrize("algorithm", const.SYMMETRIC_ALGORITHMS)
@pytest.mark.asyncio
async def test_symmetric_key_is_aliased(
    algorithm: str,
    keychain: Keychain
):
    spec = keychain.get(algorithm)
    aliased = keychain.get(spec.kid) # type: ignore
    assert spec == aliased
    assert spec.kid == aliased.kid


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_asymmetric(
    algorithm: str,
    keychain: Keychain
):
    result = keychain.filter(algorithm=algorithm)
    assert len(result) == 2


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS&const.SIGNING_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_and_use_asymmetric_sig(
    algorithm: str,
    keychain: Keychain
):
    assert len(keychain.filter(algorithm=algorithm, use='sig')) == 2
    assert len(keychain.filter(algorithm=algorithm, use='enc')) == 0


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS-const.SIGNING_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_and_use_asymmetric_enc(
    algorithm: str,
    keychain: Keychain
):
    assert len(keychain.filter(algorithm=algorithm, use='sig')) == 0
    assert len(keychain.filter(algorithm=algorithm, use='enc')) == 2


@pytest.mark.parametrize("algorithm", const.SYMMETRIC_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_symmetric(
    algorithm: str,
    keychain: Keychain
):
    result = keychain.filter(algorithm=algorithm)
    assert len(result) == 2


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS&const.DH_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_and_op_asymmetric_dh(
    algorithm: str,
    keychain: Keychain
):
    assert len(keychain.filter(algorithm=algorithm, op='decrypt')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='deriveKey')) == 2
    assert len(keychain.filter(algorithm=algorithm, op='deriveBits')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='encrypt')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='sign')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='verify')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='unwrapKey')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='wrapKey')) == 0


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS&const.ENCRYPTION_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_and_op_asymmetric_enc(
    algorithm: str,
    keychain: Keychain
):
    assert len(keychain.filter(algorithm=algorithm, op='decrypt')) == 1
    assert len(keychain.filter(algorithm=algorithm, op='deriveKey')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='deriveBits')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='encrypt')) == 2
    assert len(keychain.filter(algorithm=algorithm, op='sign')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='verify')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='unwrapKey')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='wrapKey')) == 0


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS&const.SIGNING_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_and_op_asymmetric_sig(
    algorithm: str,
    keychain: Keychain
):
    assert len(keychain.filter(algorithm=algorithm, op='decrypt')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='deriveKey')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='deriveBits')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='encrypt')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='sign')) == 1
    assert len(keychain.filter(algorithm=algorithm, op='verify')) == 2
    assert len(keychain.filter(algorithm=algorithm, op='unwrapKey')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='wrapKey')) == 0


@pytest.mark.parametrize("algorithm", const.ASYMMETRIC_ALGORITHMS&const.KEYWRAP_ALGORITHMS)
@pytest.mark.asyncio
async def test_filter_by_algorithm_and_op_asymmetric_wrap(
    algorithm: str,
    keychain: Keychain
):
    assert len(keychain.filter(algorithm=algorithm, op='decrypt')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='deriveKey')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='deriveBits')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='encrypt')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='sign')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='verify')) == 0
    assert len(keychain.filter(algorithm=algorithm, op='unwrapKey')) == 1
    assert len(keychain.filter(algorithm=algorithm, op='wrapKey')) == 2


@pytest.mark.parametrize("tags", [
    "foo",
    {"foo"},
    ["foo"]
])
@pytest.mark.asyncio
async def test_tagged(
    tags: list[str] | str | set[str],
    tagged_keys: dict[str, Any]
):
    kc = Keychain()
    kc.configure(tagged_keys)
    await kc
    assert len(kc) == len(tagged_keys) * 2
    tagged = kc.tagged(tags)
    assert len(tagged) == 4


@pytest.mark.asyncio
async def test_tagged_must_be_loaded(tagged_keys: dict[str, Any]):
    kc = Keychain()
    kc.configure(tagged_keys)
    with pytest.raises(RuntimeError):
        kc.tagged('foo')


def test_keychain_to_jwks_public(
    keychain: Keychain,
    asymmetric_keys: Sized
):
    jwks = keychain.as_jwks(private=False)
    assert isinstance(jwks, JSONWebKeySet)
    assert len(jwks) == len(asymmetric_keys)


def test_keychain_to_jwks_private(
    keychain: Keychain,
    asymmetric_keys: Sized
):
    with pytest.raises(NotImplementedError):
        keychain.as_jwks(private=True)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any

import pydantic
from cryptography.hazmat.primitives.asymmetric.types import PUBLIC_KEY_TYPES
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import load_der_public_key
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import PublicFormat
from pyasn1_modules import rfc5280

from ckms.ext import asn1
from ckms.pkix.oid import ID_EC_PUBLICKEY
from ckms.ext.pkcs.oid import RSA_PKCS1v15
from ckms.ext.rfc5480.types import ECParameters
from .subjectpublickeyalgorithmidentifier import SubjectPublicKeyAlgorithmIdentifier


class SubjectPublicKeyInfo(asn1.ASN1Model):
    __asn1spec__: type[rfc5280.SubjectPublicKeyInfo] = rfc5280.SubjectPublicKeyInfo

    algorithm: SubjectPublicKeyAlgorithmIdentifier

    public_key: asn1.fields.BitString = pydantic.Field(
        default=...,
        alias='subjectPublicKey'
    )

    @classmethod
    def frompublic(cls, key: PUBLIC_KEY_TYPES):
        return cls._frompublic(key)

    @functools.singledispatchmethod
    @classmethod
    def _frompublic(cls, key: PUBLIC_KEY_TYPES) -> 'SubjectPublicKeyInfo':
        raise TypeError(type(key).__name__)

    @_frompublic.register
    @classmethod
    def _fromec(cls, key: ec.EllipticCurvePublicKey) -> Any:
        return cls(
            algorithm=SubjectPublicKeyAlgorithmIdentifier.parse_obj({
                'algorithm': ID_EC_PUBLICKEY,
                'parameters': ECParameters.parse_obj({'namedCurve': key.curve})
            }),
            subjectPublicKey=key.public_bytes( # type: ignore
                encoding=Encoding.X962,
                format=PublicFormat.UncompressedPoint
            )
        )

    @_frompublic.register
    @classmethod
    def _fromrsa(cls, key: rsa.RSAPublicKey) -> Any:
        return cls(
            algorithm=SubjectPublicKeyAlgorithmIdentifier.parse_obj({
                'algorithm': RSA_PKCS1v15,
                'parameters': None
            }),
            subjectPublicKey=key.public_bytes( # type: ignore
                encoding=Encoding.DER,
                format=PublicFormat.PKCS1
            )
        )

    def get_public_key(self) -> PUBLIC_KEY_TYPES:
        if self.algorithm.name not in {ID_EC_PUBLICKEY, RSA_PKCS1v15}:
            raise NotImplementedError(self.algorithm)
        public_key = self.public_key
        load = load_der_public_key
        if isinstance(self.algorithm.parameters, ECParameters):
            assert self.algorithm.parameters is not None
            load = functools.partial(
                ec.EllipticCurvePublicKey.from_encoded_point,
                self.algorithm.parameters.curve.get()
            )
        return load(public_key)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from typing import Any

import pytest
from jwcrypto.jwk import JWK
from jwcrypto.jws import JWS

from ckms.core.models import KeySpecification
from ckms.core.provider.tests import JOSEConsumerTestCase
from ckms.utils import b64encode


class TestConsumeJOSE(JOSEConsumerTestCase):

    def create_jws(
        self,
        signers: list[tuple[KeySpecification, dict[str, Any] | None, dict[str, Any] | None]],
        claims: dict[str, Any] | None = None,
        compact: bool = False,
        encode: bool = False
    ) -> bytes:
        """Create a JSON Web Signature (JWS) using the given claims."""
        jws = JWS(self.payload)
        for spec, protected, header in signers:
            assert spec.algorithm is not None
            if spec.algorithm not in jws.allowed_algs: # type: ignore
                if len(signers) <= 1:
                    pytest.skip()
                continue
            key = JWK(**spec.as_jwk(private=True).dict(exclude_defaults=True))
            if compact:
                # Compact encoding must carry 'alg' in protected header
                protected = protected or {}
                protected.setdefault('alg', spec.algorithm)
            jws.add_signature( # type: ignore
                key=key,
                alg=spec.algorithm,
                protected=protected,
                header=header
            )

        buf = str.encode(cast(str, jws.serialize(compact=compact)), 'utf-8')
        if encode and not compact:
            buf = b64encode(buf)
        return buf


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from typing_extensions import Annotated

from .name import Name
from .dnsname import DNSName
from .uriname import URIName
from .ipaddressname import IPAddressName



class GeneralName(pydantic.BaseModel):
    __root__: Annotated[
        DNSName | IPAddressName | URIName,
        pydantic.Field(discriminator='kind')
    ]

    @classmethod
    def parse_obj(cls, obj: Any) -> Name:
        return super().parse_obj(obj).__root__


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
from typing import cast
from typing import Any

import pydantic
from pyasn1.type.base import Asn1Type
from pyasn1.type import univ
from pyasn1.codec.ber.decoder import decode as _decode_ber
from pyasn1.codec.ber.encoder import encode as _encode_ber
from pyasn1.codec.der.decoder import decode as _decode_der
from pyasn1.codec.der.encoder import encode as _encode_der

from . import fields


__all__: list[str] = [
    'fields',
    'ASN1Model'
]


InvalidBER: Exception = ValueError("Invalid BER encoding.")
InvalidDER: Exception = ValueError("Invalid DER encoding.")


class ObjectIdentifierEnum(enum.Enum):

    def asn1(self) -> univ.ObjectIdentifier:
        return univ.ObjectIdentifier(self.value)


class ASN1Model(pydantic.BaseModel):
    __asn1spec__: type[Asn1Type] | None = None

    @classmethod
    def parse_der(cls, value: bytes) -> 'ASN1Model':
        assert cls.__asn1spec__ is not None
        return cls.parse_obj(decode_der(value, cls.__asn1spec__()))

    def asn1(self) -> Asn1Type:
        assert self.__asn1spec__ is not None
        obj = self.__asn1spec__()
        for field in self.__fields__.values():
            v = getattr(self, field.name)
            if v is None:
                continue
            if hasattr(v, 'asn1'):
                obj[field.alias] = v.asn1() # type: ignore
            elif isinstance(v, Asn1Type):
                obj[field.alias] = v # type: ignore
        #if not obj.isValue: # type: ignore
        #    raise ValueError(
        #        f"The ASN.1 schema created by {type(self).__name__} is not valid."
        #    )
        return obj

    def der(self) -> bytes:
        return _encode_der(self.asn1()) # type: ignore

    class Config:
        arbitrary_types_allowed: bool = True


def decode_ber(
    value: bytes | Asn1Type,
    spec: Asn1Type | None = None
) -> Asn1Type:
    """Decode a BER-encoded datastructure using an optional schema."""
    if isinstance(value, univ.Any):
        value, _ = _decode_ber(bytes(value)) # type: ignore
        if _:
            raise InvalidBER
    if isinstance(value, univ.OctetString):
        value = bytes(value)
    obj, _ = _decode_ber(value, asn1Spec=spec) # type: ignore
    if _:
        raise InvalidBER
    return cast(Asn1Type, obj)


def decode_der(
    value: bytes | univ.OctetString,
    spec: Asn1Type | None = None
) -> Asn1Type:
    """Decode a DER-encoded datastructure using an optional schema."""
    obj, _ = _decode_der(value, asn1Spec=spec) # type: ignore
    if _:
        raise InvalidDER
    return cast(Asn1Type, obj)


def encode_der(
    value: bytes,
    spec: Asn1Type | None = None
) -> bytes:
    """Encode a value to bytes using an optional schema."""
    return _encode_der(value, asn1Spec=spec)


def encode_ber(
    value: Any,
    spec: Asn1Type | None = None
) -> univ.OctetString:
    """Encode a value to bytes using an optional schema."""
    return _encode_ber(value, asn1Spec=spec)


"""Declares :class:`JOSEPayloadContainer`."""
import typing


class JOSEPayloadContainer:
    """Specifies the interface for objects that can contain a payload."""
    __module__: str = 'ckms.jose'
    content_type: str | None
    token_type: str | None

    @property
    def content_type(self) -> str | None:
        return self.header.cty

    def accept(
        self,
        token_type: set[str] | None = None,
        content_type: set[str] | None = None,
    ):
        # Only perform this validation if the payload is not nested.
        if str.lower(self.content_type or '') == 'jwt':
            return

    async def get_payload(
        self,
        codec: 'PayloadCodec'
    ) -> typing.Union[bytes, str, 'JOSEPayloadContainer']:
        """Get the payload of the container object."""
        raise NotImplementedError


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from typing import cast

from ckms.types import CipherText
from ckms.types import IContentEncryptionKey
from ckms.types import IProvider
from ckms.types import PlainText
from ckms.types.encrypter import Encrypter


class ContentEncryptionKey(IContentEncryptionKey):
    __module__: str = 'ckms.core.models'

    @classmethod
    def generate(cls, algorithm: str) -> 'ContentEncryptionKey':
        """Generate a new :class:`ContentEncryptionKey`."""
        provider = IProvider.get('local')
        spec = provider.parse_spec({
            'kty': 'oct',
            'algorithm': algorithm,
        })
        return cls(spec)

    @classmethod
    def null(cls) -> 'ContentEncryptionKey':
        return cls(b'')

    async def encrypt(
        self,
        pt: bytes | PlainText,
        aad: bytes | None = None
    ) -> CipherText:
        """Encrypt the plain text with the CEK."""
        await self
        spec = cast(Encrypter, self.spec)
        ct = await spec.encrypt(pt, aad=aad)
        if inspect.isawaitable(ct):
            ct = await ct
        assert isinstance(ct, CipherText), self.spec
        return ct


    def is_aead(self) -> bool:
        return True


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Keychain`."""
import asyncio
from collections import OrderedDict
from typing import Iterator, cast
from typing import Any
from typing import Generator

from ckms.types import IKeychain
from ckms.types import IOperationPerformer
from ckms.types import JSONWebKeySet
from .models import KeySpecification
from .models import PublicKeySpecification
from .provider import Provider


class Keychain(IKeychain):
    """An in-memory :class:`~ckms.core.types.IKeychain` implementation."""
    __module__: str = 'ckms.core'
    _keys: OrderedDict[str, IOperationPerformer]

    @classmethod
    def fromdict(cls, keys: dict[str, Any]) -> 'IKeychain':
        keychain = cls() # type: ignore
        keychain.configure(keys)
        return keychain

    def __init__(
        self,
        keys: OrderedDict[str, IOperationPerformer] | None = None
    ):
        self._keys = keys or OrderedDict()

    def add(
        self,
        name: str,
        spec: KeySpecification | PublicKeySpecification | dict[str, Any]
    ) -> KeySpecification | PublicKeySpecification:
        """Add a key to the :class:`Keychain`."""
        if not isinstance(spec, (KeySpecification, PublicKeySpecification)):
            provider = Provider.get(spec.get('provider'))
            spec = cast(KeySpecification, provider.parse_spec(spec))
        if name in self._keys:
            raise ValueError(
                f"Key '{name}' is already registered in the keychain."
            )
        self._keys[name] = spec
        return spec

    def as_jwks(self, private: bool = False) -> JSONWebKeySet:
        """Return a :class:`JSONWebKeySet` instance holding the keys
        in the instance.
        """
        if private:
            raise NotImplementedError
        return JSONWebKeySet(
            keys=[
                spec.as_jwk()
                for spec in self._keys.values()
                if isinstance(spec, PublicKeySpecification)
            ]
        )

    def configure(
        self,
        keys: dict[str, dict[str, Any] | KeySpecification]
    ) -> None:
        """Configure the :class:`Keychain` from a dictionary, where the
        keys represent the names of the keys.
        """
        for name, spec in keys.items():
            self.add(name, spec)

    def dict(self) -> dict[str, Any]: # pragma: no cover
        """Return the keys in the :class:`Keychain` as a dictionary, mapping
        key identifiers to key objects.
        """
        return OrderedDict(self._keys)

    def filter(
        self,
        algorithm: str | None = None,
        kid: str | None = None,
        use: str | None = None,
        op: str | None = None
    ) -> 'IKeychain':
        Impl = type(self)
        if kid is not None and not self.has(kid):
            return Impl(keys=OrderedDict())
        keys: OrderedDict[str, IOperationPerformer] = OrderedDict()
        for name, spec in self._keys.items():
            spec = cast(KeySpecification, spec)
            if not spec.is_loaded():
                raise RuntimeError("Specification is not loaded.")
            assert spec.algorithm is not None
            if algorithm is not None and spec.algorithm != algorithm:
                continue
            if use is not None and spec.use != use:
                continue
            if op is not None and not spec.can_perform(op):
                continue
            keys[name] = spec 
        return Impl(keys=keys)

    def get(self, kid: str | None) -> KeySpecification:
        if not self.has(kid):
            raise KeyError(kid)
        return cast(KeySpecification, self._keys[kid or ''])

    def has(self, kid: str | None) -> bool:
        return kid in self._keys

    def private(self) -> 'IKeychain':
        """Return the :class:`Keychain` with only the keys that can perform
        private operations.
        """
        return type(self)(keys={
            x.kid: x for x in self._keys.values()
            if isinstance(x, KeySpecification)
        })

    def public(self) -> 'IKeychain':
        """Return the :class:`Keychain` with only the keys that can perform
        public operations.
        """
        return type(self)(keys={
            x.kid: x for x in self._keys.values()
            if isinstance(x, PublicKeySpecification)
        })

    def tagged(self, tags: str | list[str] | set[str]) -> 'IKeychain':
        Impl = type(self)
        if isinstance(tags, str):
            tags = {tags}
        if not isinstance(tags, set):
            tags = set(tags)
        keys: OrderedDict[str, IOperationPerformer] = OrderedDict()
        for name, spec in self._keys.items():
            if not bool(spec.tags & tags):
                continue
            if not spec.is_loaded():
                raise RuntimeError(
                    "Keys must be loaded before creating a new instance."
                )
            keys[name] = spec

        return Impl(keys=keys)

    async def _configure(self) -> IKeychain:
        await asyncio.gather(*[self._load_spec(x) for x in self])
        return self

    async def _load_spec(self, spec: KeySpecification) -> KeySpecification:
        if not spec.is_loaded():
            await spec
            assert spec.kid is not None
            if spec.is_asymmetric() and not spec.is_public():
                assert spec.kid is not None
                self._keys[spec.kid] = spec.as_public()
            elif spec.kid not in self._keys:
                self._keys[spec.kid] = spec
        return spec

    def __await__(self) -> Generator[Any, None, IKeychain]:
        return self._configure().__await__()

    def __iter__(self) -> Iterator[KeySpecification]:
        return iter(list(self._keys.values())) # type: ignore

    def __len__(self) -> int:
        return len(self._keys)


default: Keychain = Keychain()


def get_default_keychain() -> Keychain: # pragma: no cover
    """Return the default :class:`Keychain` instance used by the
    application.
    """
    return default



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from httpx import AsyncClient

from ckms.ext.tsa import TimestampingAuthority
from ckms.ext.tsa.types import TimeStampResp


__all__: list[str] = [
    'test_create_timestamp_from_bytes'
]


@pytest.mark.asyncio
async def test_create_timestamp_from_bytes(
    client: AsyncClient,
    tsa: TimestampingAuthority,
    data: bytes,
    digest: bytes
):
    response = await tsa.timestamp(
        client=client,
        data=data,
        digest='sha256'
    )
    assert isinstance(response, TimeStampResp)
    info = response.get_tst_info()
    assert info is not None, response
    assert response.token is not None
    response.token.verify(tsa.certificate)


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from ckms.types import JSONWebKey



@pytest.mark.parametrize("alg,use", [
    ("RS256", "sig"),
    ("RSA-OAEP", "enc"),
])
@pytest.mark.parametrize("path,public,asymmetric", [
    ('pki/rsa.key', False, True),
    ('pki/rsa.pub', True, True),
])
def test_frompem(path: bytes, alg: str, use: str, public: bool, asymmetric: bool):
    pem = open(path, 'rb').read()
    jwk = JSONWebKey.frompem(
        pem,
        alg=alg,
        use=use
    )
    assert jwk.is_public() is public
    assert jwk.is_asymmetric() is asymmetric
    assert jwk.alg is not None
    assert jwk.use is not None


def test_as_public_from_private():
    pem = open('pki/rsa.key', 'rb').read()
    private = JSONWebKey.frompem(
        pem,
        alg='RS256',
        use='sig'
    )
    public = private.as_public()
    p1 = public.get_public_key()
    p2 = private.get_public_key()
    assert private != public
    assert p1.public_numbers() == p2.public_numbers()


def test_as_public_from_public():
    pem = open('pki/rsa.pub', 'rb').read()
    private = JSONWebKey.frompem(
        pem,
        alg='RS256',
        use='sig'
    )
    public = private.as_public()
    assert private == public



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
"""Declares :class:`BaseKeySpecification`."""
from typing import Any

import pydantic

from ckms import core
from ckms.types import IKeySpecification
from ckms.types import IProvider


class BaseKeySpecification(pydantic.BaseModel, IKeySpecification):
    """The base class for all provider key configurations."""
    __module__: str = 'ckms.core.types'

    @pydantic.validator('provider', allow_reuse=True, check_fields=False)
    def _validate_provider(cls, value: str | IProvider) -> IProvider:
        assert value is not None # nosec
        assert value != 'none'
        if isinstance(value, str): # pragma: no cover
            value = core.provider(value)
        return value

    def can_decrypt(self) -> bool:
        """Return a boolean indicating if the key may be used to
        perform encryption operations.
        """
        return self.use == 'enc'

    def can_encrypt(self) -> bool:
        """Return a boolean indicating if the key may be used to
        perform encryption operations.
        """
        return self.use == 'enc'

    def can_sign(self) -> bool:
        """Return a boolean indicating if the key may be used to
        create digital signatures.
        """
        return self.use == "sig"

    def is_asymmetric(self) -> bool:
        """Return a boolean indicating if the key is an asymmetric
        algorithm.
        """
        return str(self.kty) in {"EC", "OKP", "RSA"}

    def is_symmetric(self) -> bool:
        """Return a boolean indicating if the key is a symmetric
        algorithm.
        """
        return self.kty in {"oct"}

    def is_loaded(self) -> bool: # pragma: no cover
        """Return a boolean indicating if the metadata describing the
        key is retrieved.
        """
        return self.loaded

    def is_public(self) -> bool:
        """Return a boolean indicating if this specification represent
        a public key.
        """
        return False

    class Config:
        arbitrary_types_allowed: bool = True


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`GenerateKeySpec`."""
from typing import Any
from typing import Literal
from typing import TypeAlias
from typing_extensions import Annotated

import pydantic

from .edwardscurvetype import EdwardsCurveType
from .ellipticcurvetype import EllipticCurveType


__all__: list[str] = [
    'GenerateAES',
    'GenerateEllipticCurve',
    'GenerateHMAC',
    'GenerateKeySpec',
    'GenerateOKP',
    'GenerateRSA',
    'KeyType'
]


class GenerateRSA(pydantic.BaseModel):
    kty: Literal['RSA'] = 'RSA'
    length: int


class GenerateOKP(pydantic.BaseModel):
    kty: Literal['OKP'] = 'OKP'
    curve: EdwardsCurveType


class GenerateEllipticCurve(pydantic.BaseModel):
    kty: Literal['EC'] = 'EC'
    curve: EllipticCurveType


class GenerateAES(pydantic.BaseModel):
    kty: Literal['oct'] = 'oct'
    length: int
    use: Literal['enc']


class GenerateHMAC(pydantic.BaseModel):
    kty: Literal['oct'] = 'oct'
    length: int
    use: Literal['sig']


GenerateSymmetric = Annotated[
    GenerateHMAC|GenerateAES,
    pydantic.Field(discriminator='use')
]

KeyType: TypeAlias = (
    GenerateRSA |
    GenerateOKP |
    GenerateEllipticCurve |
    GenerateSymmetric
)


class GenerateKeySpec(pydantic.BaseModel):
    __root__: Annotated[
        KeyType,
        pydantic.Field(discriminator='kty')
    ]

    @classmethod
    def parse_spec(
        cls,
        spec: dict[str, Any]
    ) -> KeyType:
        """Parse the spec as the appropriate key type."""
        return cls.parse_obj(spec).__root__


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`EdwardsCurveAlgorithmType`."""
import enum


class EdwardsCurveAlgorithmType(str, enum.Enum):
    EdDSA           = 'EdDSA'
    ECDH_ES         = 'ECDH-ES'
    ECDH_ES_A128KW  = 'ECDH-ES+A128KW'
    ECDH_ES_A192KW  = 'ECDH-ES+A192KW'
    ECDH_ES_A256KW  = 'ECDH-ES+A256KW'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unimatrix.exceptions import CanonicalException


__all__: list[str] = [
    'MaximumAgeExceeded',
    'TokenExpired',
    'TokenNotEffective',
    'WrongAudience',
]


class InvalidToken(CanonicalException):
    http_status_code: int  = 403
    code: str = "TOKEN_INVALID"
    message: str = "The supplied JSON Web Token (JWT) is not valid."
    hint: str = "Consult the documentation for the required claims and their content."


class TokenExpired(InvalidToken):
    code: str = "TOKEN_EXPIRED"
    detail: str = (
        "The 'exp' claim specified a fatal term, after which the supplied "
        "token must not be used."
    )


class TokenNotEffective(InvalidToken):
    code: str = 'TOKEN_NOT_EFFECTIVE'
    detail: str = (
        "The 'nbf' claim specified a date and time before which the token "
        "must not be accepted, and it lies in the future."
    )


class WrongAudience(InvalidToken):
    code: str = "WRONG_AUDIENCE"
    detail: str = (
        "The audience(s) specified by the 'aud' claim are not "
        "accepted by the application."
    )


class WrongIssuer(InvalidToken):
    code: str = 'WRONG_ISSUER'
    detail: str = (
        "The issuer specified by the 'iss' claim is not "
        "accepted by the application."
    )


class MaximumAgeExceeded(TokenExpired):

    def __init__(self, max_age: int, now: int, issued: int):
        super().__init__(
            detail=(
                "The JSON Web Token (JWT) exceeded the maximum age of "
                f"{max_age} seconds with {now-issued}."
            )
        )


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from cryptography.hazmat.primitives.asymmetric import ec

from ckms.ext.asn1 import ObjectIdentifierEnum


class EllipticCurveType(str, ObjectIdentifierEnum):
    SECP256R1 = '1.2.840.10045.3.1.7'
    SECP256K1 = '1.3.132.0.10'
    SECP384R1 = '1.3.132.0.34'
    SECP521R1 = '1.3.132.0.35'

    @classmethod
    def fromcurve(cls, curve: ec.EllipticCurve) -> 'EllipticCurveType':
        return getattr(cls, str.upper(curve.name))

    def get(self) -> ec.EllipticCurve:
        return getattr(ec, self.name)()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic

from ckms import types
from .localkeyspecification import LocalKeySpecification


class TransientKey(pydantic.BaseModel):
    length: int

    async def setup(self, spec: LocalKeySpecification) -> Any:
        assert isinstance(spec.provider, types.IProvider) # nosec
        return await spec.provider.generate(
            types.GenerateKeyOperation(
                kty=spec.kty,
                length=self.length,
                curve=spec.curve,
                use=spec.use
            )
        )


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .asymmetricsigningtestcase import AsymmetricSigningTestCase
from .baseprovidertestcase import BaseProviderTestCase
from .encryptiontestcase import EncryptionTestCase
from .josebasetestcase import JOSEBaseTestCase
from .joseconsumertestcase import JOSEConsumerTestCase
from .joseencryptiontestcase import JOSEEncryptionTestCase
from .joseproducertestcase import JOSEProducerTestCase
from .providertestcase import ProviderTestCase
from .symmetricsigningtestcase import SymmetricSigningTestCase


__all__: list[str] = [
    'AsymmetricSigningTestCase',
    'BaseProviderTestCase',
    'EncryptionTestCase',
    'JOSEBaseTestCase',
    'JOSEConsumerTestCase',
    'JOSEEncryptionTestCase',
    'JOSEProducerTestCase',
    'ProviderTestCase',
    'SymmetricSigningTestCase'
]


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from pyasn1_modules import rfc5280

from .attribute import Attribute


class RelativeDistinguishedName(list[Attribute]):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: list[Attribute] | rfc5280.RelativeDistinguishedName | None
    ) -> Optional['RelativeDistinguishedName']:
        if value is None:
            return None
        if isinstance(value, rfc5280.RelativeDistinguishedName):
            value = [
                Attribute.parse_obj({'type': str(x[0]), 'value': x[1]}) # type: ignore
                for x in value # type: ignore
            ]
        assert isinstance(value, list)
        return cls(value)

    def asn1(self) -> rfc5280.RelativeDistinguishedName:
        obj = rfc5280.RelativeDistinguishedName()
        for attr in self:
            obj.append(attr.asn1()) # type: ignore
        return obj


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
import json

import pytest

from ckms.core.models import JOSEObject
from ckms.core.models.jsonwebsignature import JSONWebSignature


def test_deserialize_jws_json(multi_jws: bytes):
    obj = JOSEObject.parse(multi_jws)
    assert isinstance(obj, JSONWebSignature)


# pylint: skip-file
import pytest

from .conftest import SIGNING_ALGORITHMS


@pytest.mark.parametrize("using,algorithm", SIGNING_ALGORITHMS)
class TestSigning:

	@pytest.mark.asyncio
	async def test_sign(self, keychain, data, using, algorithm):
		sig = await keychain.sign(data=data, using=using, algorithm=algorithm)
		assert await keychain.verify(signature=sig, data=data, using=using, algorithm=algorithm)



"""Declares :class:`JSONWebEncryptionRecipient`."""
import typing

import pydantic

from .octetb64 import OctetB64


class JSONWebEncryptionRecipient(pydantic.BaseModel):
    __module__: str = 'ckms.core.models'
    header: dict[str, typing.Any] = {}
    encrypted_key: OctetB64 | None = None


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Decoder`."""
import inspect
import functools
from typing import Any

from ckms.core.models import JOSEObject
from ckms.core.models.jsonwebsignature import JSONWebSignature
from ckms.core.models.jsonwebencryption_ import JSONWebEncryption
from ckms.utils import b64decode
from ckms.types import IKeychain
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload
from .exceptions import InvalidSignature
from .joseheaderset import JOSEHeaderSet


class Decoder:
    """Providers an interface to decrypt and introspect JOSE objects."""
    __module__: str = 'ckms.jose'
    model: type[JOSEObject] = JOSEObject
    decrypter: IKeychain
    verifier: IKeychain
    jwt_types: set[str]

    def __init__(
        self,
        decrypter: IKeychain,
        verifier: IKeychain,
        jwt_types: set[str] | None = None
    ):
        self.decrypter = decrypter
        self.jwt_types = jwt_types or {"at+jwt", "jwt", "secevent+jwt"}
        self.verifier = verifier

    async def decode(
        self,
        serialized: bytes | str,
        verify: bool = True,
        require_kid: bool = True,
        accept: set[str] = {"at+jwt", "jwt", "jwt+secevent"},
        **kwargs: Any
    ):
        """Decode the given JOSE object `obj`, which is either JSON Web
        Encryption (JWE) or a JSON Web Signature (JWS) using any supported
        serialization format.

        The `verify` parameter indicates if the signatures must validate
        using the configured :class:`ckms.types.IKeychain` instance. By
        default, all signatures must validate with a key that was
        preregistered in the keychain.

        The :meth:`decode()` method requires all signers to provide a
        ``kid`` header parameter used to lookup the corresponding (public)
        key from the keychain. This behavior may be changed by providing
        the `require_kid` parameter. If `require_kid` is ``False``, then
        :class:`Decoder` attempts all keys in :attr:`Decoder.keychain`
        that match the signature algorithm. The default value is ``True``.
        """
        obj = await self._decode(self.model.parse(serialized))
        if isinstance(obj, JSONWebSignature):
            if verify\
            and not await obj.verify(self.verifier, require_kid=require_kid):
                raise InvalidSignature
            typ = obj.get_type()
            payload = obj.get_payload()

            # TODO: Implement a parser system like in cbra
            if accept:
                payload = obj.claims(accept=accept)
            elif typ is not None and typ in self.jwt_types:
                payload = JSONWebToken.frompayload(payload)
            else:
                payload = b64decode(payload)
        elif isinstance(obj, JSONWebToken):
            payload = obj
        else:
            payload = obj
        return payload

    async def jws(self, token: bytes | str) -> JSONWebSignature:
        """Like :meth:`decode()`, but returns the unverified and unparsed
        JSON Web Signature (JWS). If `token` is not a JWS, raise an
        exception.
        """
        obj = await self._decode(self.model.parse(token))
        if not isinstance(obj, JSONWebSignature):
            raise MalformedPayload(
                detail=(
                    "The applications expects a JSON Web Signature (JWS)."
                )
            )
        return obj

    async def jwt(
        self,
        token: bytes | str,
        accept: set[str] | str | None = None
    ) -> tuple[JSONWebSignature, JSONWebToken]:
        """Like :meth:`decode()`, but returns a tuple containing the
        unverified JSON Web Signature (JWS) and its claims set. If
        `token` is not a JWS, or does not have a claims set, raise
        an exception.
        """
        if isinstance(accept, str):
            accept = {accept}
        sig = await self.jws(token)
        return sig, sig.claims(accept or self.jwt_types)

    def introspect(
        self,
        value: str | bytes,
        model: type[JSONWebToken] = JSONWebToken,
        accept: set[str] = {"at+jwt", "secevent+jwt", "jwt"}
    ) -> tuple[JOSEHeaderSet, JSONWebToken | None]:
        """Introspect a JOSE object and return the list of :class:`JOSEHeader`
        instances for all recipients or signatures.
        """
        obj = JOSEObject.parse(value)
        claims = None
        if isinstance(obj, JSONWebSignature) and obj.is_jwt(accept=accept):
            claims = obj.get_claims(model)
        return (JOSEHeaderSet(headers=obj.get_headers()), claims)

    @functools.singledispatchmethod
    def _decode(self, obj: Any) -> Any:
        raise NotImplementedError(repr(obj))

    @_decode.register
    async def _decode_jwe(self, obj: JSONWebEncryption) -> bytes | JSONWebToken | JSONWebSignature:
        cty = obj.get_content_type()
        typ = obj.get_type()
        payload = obj.decrypt(self.decrypter)
        if inspect.isawaitable(payload):
            payload = await payload
        if cty is None and typ in self.jwt_types:
            # JWE with no further content type specification. If this is JWT,
            # then the JWE contains an unsigned JSON Web Token (JWT). Otherwise,
            # we can not determine the payload content type and thus return the
            # raw byte-sequence.
            payload = JSONWebToken.fromjson(payload)
        elif cty in {"jose", "jose+json", *self.jwt_types}:
            # Nested JWT -> In the case that nested signing or
            # encryption is employed, this Header Parameter MUST
            # be present (RFC 7519, 5.2).
            payload = JOSEObject.parse(payload)
            assert isinstance(payload, JSONWebSignature)
        return payload

    @_decode.register
    async def _decode_jws(self, obj: JSONWebSignature) -> JSONWebSignature:
        return obj


import os

from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import GCM
from google.cloud import kms



algorithm = AES(b'0' * 32)



def encrypt_local(pt: bytes) -> tuple[bytes, bytes]:
    iv = os.urandom(12)
    cipher = Cipher(algorithm, GCM(iv))
    e = cipher.encryptor()
    ct = e.update(pt) + e.finalize()
    return (ct, iv)



if __name__ == '__main__':
    breakpoint()



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from typing import Any
from typing import Generator
from typing import Iterator
from typing import TypeAlias

from .decrypter import Decrypter
from .ioperationperformer import IOperationPerformer
from .verifier import Verifier


KeyType: TypeAlias = Decrypter | Verifier


class IKeychain:
    __module__: str = 'ckms.types'

    @classmethod
    def fromdict(cls, keys: dict[str, Any]) -> 'IKeychain':
        raise NotImplementedError

    def as_jwks(self, private: bool = False) -> Any:
        """Return a :class:`JSONWebKeySet` instance holding the keys
        in the instance.
        """
        raise NotImplementedError

    def dict(self) -> dict[str, Any]:
        raise NotImplementedError

    def filter(
        self,
        algorithm: str | None = None,
        kid: str | None = None,
        use: str | None = None,
        op: str | None = None,
        keys: Any = None
    ) -> Any:
        for spec in keys:
            spec = cast(IOperationPerformer, spec)
            if not spec.is_loaded():
                raise RuntimeError("Specification is not loaded.")
            assert spec.algorithm is not None
            if algorithm is not None and spec.algorithm != algorithm:
                continue
            if use is not None and spec.use != use:
                continue
            if op is not None and not spec.can_perform(op):
                continue
        return keys

    def get(self, kid: str | None) -> KeyType:
        """Lookup a key by its key identifier."""
        raise NotImplementedError

    def tagged(self, tags: str | list[str] | set[str]) -> 'IKeychain':
        """Return a new :class:`IKeychain` instance with the keys
        matching the given tags.
        """
        raise NotImplementedError

    def has(self, kid: str | None) -> bool:
        """Return a boolean indicating if the given key identifier is in
        the keychain.
        """
        raise NotImplementedError

    def verifiers(self) -> Iterator[Verifier]:
        """Return all keys that can be used to verify a digital
        signature.
        """
        for x in self:
            if not x.can_verify():
                continue
            assert isinstance(x, Verifier)
            yield x

    def __and__(self, keychain: 'IKeychain') -> 'IKeychain':
        return self.fromdict({
            **self.dict(),
            **keychain.dict()
        })

    def __await__(self) -> Generator[Any, None, 'IKeychain']:
        raise NotImplementedError

    def __contains__(self, kid: str) -> bool:
        raise NotImplementedError

    def __iter__(self) -> Iterator[KeyType]:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import random
from typing import cast
from typing import Any
from typing import Awaitable

from ckms.types import Digest
from ckms.types import Message
from ckms.types import IHTTPClient
from .timestampingauthority import TimestampingAuthority
from .types import TimeStampResp


class Timestamper:
    """Provides an interface to retrieve a timestamp of a digest
    using multiple :rfc:`3161` timestamping CAs.
    """
    __module__: str = 'ckms.ext.tsa'
    servers: list[TimestampingAuthority]

    @staticmethod
    def default() -> 'Timestamper':
        return _default

    @classmethod
    def fromlist(cls, authorities: list[dict[str, Any]]) -> 'Timestamper':
        return cls(list(TimestampingAuthority(**x) for x in authorities))

    def __init__(self, servers: list[TimestampingAuthority]):
        self.servers = servers

    def select(self, n: int) -> list[TimestampingAuthority]:
        """Select the given number of :class:`TimestampingAuthority`
        instances from the preconfigured list.
        """
        selected: list[TimestampingAuthority] = []
        candidates = list(self.servers)
        while candidates:
            i = random.choice(range(len(candidates)))
            selected.append(candidates.pop(i))
            if len(selected) == n:
                break
        return selected

    async def timestamp(
        self,
        client: Any,
        data: bytes | Digest | Message,
        digestmod: str = 'sha256',
        min_signatures: int = 1,
        retry: int = 0,
    ) -> list[TimeStampResp]:
        """Return a list of :rfc:`3161` timestamps, represented as :class:`Timestamp`
        instances, for the given input data `data` using hashing algorithm `digestmod`.

        The `min_signatures` parameter specifies how many signatures must be
        obtained for the input data. The output is not guaranteed to contain `n`
        signatures; the `min_signatures` parameter solely indicates the number
        of authorities that should be attempted.
        """
        if isinstance(data, Digest) and data.digestmod != digestmod:
            assert data.digestmod is not None
            digestmod = data.digestmod
        client = cast(IHTTPClient, client)
        futures: list[Awaitable[TimeStampResp]] = []
        for tsa in self.select(min_signatures):
            fut = tsa.timestamp(
                client=client,
                data=data,
                digest=digestmod
            )
            futures.append(fut)
        return cast(list[TimeStampResp], await asyncio.gather(*futures))


_default: Timestamper = Timestamper([])